
/* This code is Copyright(C) Paul Jimenez, 1997, but completely freely redistributable */

package org.place.pj;
import java.util.*;

/** A class that does approximately what getopt() does in C.
  * For use in Java applications.  Now with longopt support!
  *
  * @version 2.3, 3/31/98
  * In response to suggestions and a patch from David Karr <dkarr@nmo.gtegsc.com>
  *  uch-belated fix to problem of distinguishing lack of args from
  * empty-string arg.   Also fixed another minor bug reported by 
  * Vadim Zaliva <Vadim.Zaliva@us.matranet.com>.  Oh, and packaged up
  * the whole thing into org.place.pj.
  *
  * @version 2.2, 9/24/97
  * Added 'required options' constructor - options in said params
  * MUST be provided or else we throw a MissingCommandlineArgument
  * exception.
  *
  * @version 2.1, 9/24/97
  * In response to a bug report by Mike Powers <mpowers@teir.com>
  * fixed a bug in option-parameter handling where if an option was
  * specified without a parameter it would skip the next arg in
  * the list
  *
  * @version 2.0, 6/10/97
  * Added GNU-style longopt support, which is enough of a change to
  * warrant a full-version increment.  Rewrote the whole constructor
  * from scratch.
  *
  * @version 1.1, 3/19/97
  * In response to a bug report by Ed Wagner <ewagner@n-sight.com>, 
  * removed line 'import java.lang.*' and changed the optionlist 
  * Hashtable to initialize to size 1 instead of validflags.length()
  * because the latter would fail if no flags were specified.
  * 
  * @author Paul Jimenez
  * @version 1.0, 1/5/97
  * Initial release.
  * 
  */

public class Commandline extends Object {

  /** The list of parameters that weren't arguments to options */
  public String params[];

  /** The internal storage of options */
  Hashtable optionlist; 

  /** debugging flag - enable this to send debug output to System.out */ 
  public boolean debug = false;


  /**   we only want work hard once, so we do all the parsing here
    *
    * @param argv[] array of strings (usually argv)
    * @param validflags a string of valid options that don't take arguments
    * @param validoptions a string of valid options that take (optional) arguments
    * @param valindlongflags[] array of strings that are valid options that don't 
       take arguments
    * @param valindlongoptions[] array of strings that are valid options that 
       take (optional) arguments
    * @exception InvalidCommandlineArgument iff an invalid argument is 
       found on the commandline
    */  

  public Commandline(String argv[], 
                     String validflags, String validoptions,
                     String validlongflags[], String validlongoptions[]) throws InvalidCommandlineArgument {

    doParse(argv, validflags, validoptions, validlongflags, validlongoptions);
  }

  /** for backwards compatibility and those times when you don't 
    * <EM>want</EM> longopts 
    * @param argv[] array of strings (usually argv)
    * @param validflags a string of valid options that don't take arguments
    * @param validoptions a string of valid options that take (optional) arguments
    * @exception InvalidCommandlineArgument iff an invalid argument is 
       found on the commandline
    */ 

  public Commandline(String argv[], 
                     String validflags, 
		     String validoptions) throws InvalidCommandlineArgument {
    String nullarray[] = {};
    doParse(argv, validflags, validoptions, nullarray, nullarray);
  }

  /** with required options that MUST have parameters if specified 
    * 
    * @param argv[] arry of strings (usually argv)
    * @param validflags a string of valid options that don't take arguments
    * @param validoptions a string of valid options that take (optional) arguments
    * @param requiredoptions a string of valid options that must be provided with arguments
    * @param validlongflags[] array of strings that are valid options that don't
        take arguments
    * @param validlongoptions[] array of strings that are valid options that take
        (optional) arguments
    * @param requiredlongoptions[] array of strings that are valid options that must 
        be provided with arguments
    * @exception InvalidCommandlineArgument iff an invalid argument is found on the commandline
    * @exception MissingCommandlineArguement iff a required option is specified without an option
  */
  
  public Commandline(String argv[], 
                     String validflags, String validoptions, String requiredoptions, 
                     String validlongflags[], String validlongoptions[], String requiredlongoptions[] ) 
			throws InvalidCommandlineArgument, MissingCommandlineArgument {

    boolean missedsome = false;
    String missed = "Missing Commandline Option(s): ";
    String tempvalidoptions[] = new String[validlongoptions.length + requiredlongoptions.length];
    int i, j;
  
    /* tempvalidoptions = validlongoptions+requiredlongoptions */ 
    i = j = 0;
    while(i<validlongoptions.length){
      tempvalidoptions[j] = validlongoptions[i];
      i++; j++;
    }
    i = 0;
    while (i<requiredlongoptions.length) {
      tempvalidoptions[j] = requiredlongoptions[i];
      i++; j++;
    }
    /* call and parse normally */
    doParse(argv, validflags, validoptions + requiredoptions, validlongflags, tempvalidoptions);

    /* check for requireds that are missing */ 
    for(i = 0; i < requiredoptions.length(); i++) {
      if (!hasOption(new Character(requiredoptions.charAt(i)))) {
        missed = missed + " " + (new Character(requiredoptions.charAt(i)).toString());
        missedsome = true;
      }
    }

    for(i = 0; i < requiredlongoptions.length; i++) {
      if (!hasOption(requiredlongoptions[i])) {
        missed = missed + " " + requiredlongoptions[i];
        missedsome = true;
      }
    } 

    /* if any are missing, throw */
    if (missedsome) {
      throw new MissingCommandlineArgument(missed);
    } 
     
  }


  void doParse(String argv[], 
               String validflags, String validoptions,
               String validlongflags[], String validlongoptions[]) throws InvalidCommandlineArgument {

    int i, o;
    boolean shouldthrow = false;
    String throwstring = new String("Invalid Commandline argument(s): ");
    Hashtable possopts = new Hashtable();
    String arg, optarg;
    boolean moreargs = true;     
    Vector paramv = new Vector(1,1);

    optionlist = new Hashtable(1,1);
 
    /* keep this O(lenght(possible options) + length(arguments) by pulling possopts into a hashtable */

    for(i = 0; i <validlongflags.length ; i++) {
      if (debug) {
        System.out.println("debug: Looking for arg \""+validlongflags[i]+"\"");
      }
      possopts.put(validlongflags[i], Boolean.FALSE);
    }
    for(i = 0; i <validlongoptions.length ; i++) {
      if (debug) {
        System.out.println("debug: Looking for arg \""+validlongoptions[i]+"\"");
      }
      possopts.put(validlongoptions[i], Boolean.TRUE);
    }
    for(i = 0; i <validflags.length() ; i++) {
      if (debug) {
        System.out.println("debug: Looking for arg "+(new Character(validflags.charAt(i))).toString());
      }
      possopts.put(new Character(validflags.charAt(i)).toString(), Boolean.FALSE);
    }
    for(i = 0; i <validoptions.length() ; i++) {
      if (debug) {
        System.out.println("debug: Looking for arg "+(new Character(validoptions.charAt(i))).toString());
      }
      possopts.put(new Character(validoptions.charAt(i)).toString(), Boolean.TRUE);
    }

    /* now that we're initialized, run thru the args */

    i = 0;
    while (i<argv.length) {
      arg = argv[i];
      if (debug) {
        System.out.println("debug: Arg \""+arg+"\": ");
      }
      if (moreargs) { /* haven't hit 'end of args' separator yet */
        if (arg.equalsIgnoreCase("--")) { /* special case */
          moreargs = false;
        } else if (islongopt(arg)) { /* it's a long option */
          o = 0;
          arg = arg.substring(2);
          if (debug) { 
            System.out.println("debug:   looking for internal flag \""+arg+"\"");
          } 
          if (possopts.containsKey(arg)) { /* valid one, too */
            if (possopts.get(arg) == Boolean.FALSE) { /* it's a flag */
              sethasflag(arg);
            } else { /* it's an option */
              o++;
              sethasoption(arg, optarg = paramfor(argv, i, o));
              if (optarg == null ) {
                o--;
              } 
            }
          } else { /* invalid arg, add to throw */ 
            throwstring = throwstring + " " +arg; 
            shouldthrow = true;
          }
          i = i + o; 
        } else if (isoptlist(arg)) { /* it's an option list (shortopts) */
          o = 0;
          for (int j = 1; j <arg.length(); j++) { /* cycle thru the shortopts */
            String charatj = new Character(arg.charAt(j)).toString(); /* get each one individually */
            if (possopts.containsKey(charatj)) { /* valid arg */
              if (possopts.get(charatj) == Boolean.FALSE) { /* it's a flag */
                sethasflag(charatj);
              } else { /* it's an option */
                o++;
                sethasoption(charatj, optarg = paramfor(argv, i, o));
                if (optarg == null) {
                  o--;
                } 
              }
            } else { /* invalid arg, add to throw */
              throwstring = throwstring +" "+charatj; 
              shouldthrow = true;
            } 
          } 
          i = i + o;  /* skip params to options */
        } else { /* it's a param */
          paramv.addElement(arg);
          if (debug) {
            System.out.println("debug:   has param "+arg);
          }
        } 
      } else { /* the rest of the args are params */ 
        paramv.addElement(arg);
        if (debug) {
          System.out.println("debug:   has param "+arg);
        }
      }
      i++;
    }
    params = new String[paramv.size()];      
    paramv.copyInto(params);
    if (shouldthrow) {
      throw new InvalidCommandlineArgument(throwstring);
    }
  }
     
  String paramfor(String arglist[], int optat, int argat) {
    for(int i=optat+1; i <= optat+argat; i++) {
      if ((i >= arglist.length) || (isoptlist(arglist[i]))) {
        /* if we run off the end of the list or hit an optlist on the way to the argat'th */
        if (debug) {
          System.out.print("debug:   escaping paramfor early");
          if (i >= arglist.length) {
            System.out.println(" cuz arglist is too short.");
          } else {
            System.out.println(" cuz we hit a new opt list.");
          }
        }
        return null;
      }
    }
    return arglist[optat+argat];
  }

  protected class StringHolder {
    public String contents;
    public StringHolder(String in) {
      contents = in;
    }
  }

  synchronized void sethasflag(String s) {
    if (debug) {
      System.out.println("debug:   has flag "+s);
    }
    optionlist.put(s,new StringHolder(null));  
  }

  synchronized void sethasoption(String o, String p) {
    if (debug) {
      System.out.println("debug:   has option "+o+" with value "+p);
    }
    optionlist.put(o,new StringHolder(p));  
  }

  boolean isoptlist(String s) {
    if (debug) {
      System.out.println("debug:   calling isoptlist("+s+")");
    }
    return ((s.length() > 1) && (s.charAt(0) == '-'));
  }

  boolean islongopt(String s) {
    return ((s.length() >2) && (s.charAt(0) == '-') && (s.charAt(1) == '-'));
  }

  /* The below, together with params[], is the entire API :) */

  public boolean hasOption(Character option) {
    return hasOption(option.toString());
  }

  /**
    *whether or not the commandline contains a specified option 
    *@param option a single-character string of the option you want to check on
    *@return true if option was specified on the command line or
      false if option was NOT specified on the command line
  */
  public boolean hasOption(String option) {
    return optionlist.containsKey(option);
  }

  public String getOption(Character option) {
    return getOption(option.toString());
  }
 
  /**
    *the 'value' of an option
    *@param option a string (single-character for short options)
      of the option you want returned
    *@return the argument specified for option on the commandline,
      or null if the option wasn't specified 
  */
  public String getOption(String option) { 
    return ((StringHolder)optionlist.get(option)).contents;
  }
 

  /* Integrated Test Tools */
  /** A test routine that I used to verify everything was working correctly.
      It sends output to STDOUT, when called from a program like
<PRE>
import Commandline;

class test {

  public static void main(String argv[]) {

    String in[] = { "-ac", "c_arg", "param-1", "--flaga", "--opta", "opta_arg" , "--", "-param-3", "param-4" };
    String longflags[] = { "flaga", "flagb" };
    String longopts[] = { "opta", "optb" };

    Commandline.test(in, "ab", "cdef", longflags, longopts);
  }

}
</PRE>
    which should output 
<PRE>

Valid short options are: cdef (with args) and ab (without args).
Valid long options are:
  opta
  optb
Valid long flags are:
  flaga
  flagb
Running with options: -ac c_arg param-1 --flaga --opta opta_arg -- -param-3 param-4  ...
... found option c with value c_arg
... didn't find option d
... didn't find option e
... didn't find option f
... found option a with value
... didn't find option b
... found option opta with value opta_arg
... didn't find option optb
... found option flaga with value
... didn't find option flagb
... Extra Parameters: param-1 -param-3 param-4
Done.
</PRE> 

  * @param in an array of strings which are arguments
  * @param flags a string of options that don't take arguments
  * @param opts a string of options that take (optional) arguments
  */
  public static void test(String in[], String flags, String opts, String longflags[], String longopts[]) {

    String option;

    /* show what we're testing */ 
    System.out.println("Valid short options are: "+opts+" (with args) and "+flags+" (without args).");
    System.out.println("Valid long options are: ");
    for (int i=0; i<longopts.length; i++) {
      System.out.println("  "+longopts[i]);
    }
    System.out.println("Valid long flags are: ");
    for (int i=0; i<longflags.length; i++) {
      System.out.println("  "+longflags[i]);
    }
    System.out.print("Running with options: ");
    for (int i=0; i<in.length; i++) {
      System.out.print(in[i]+" ");
    }
    System.out.println(" ...");
  
    /* run the tests */
    
    try { 
      Commandline c = new Commandline(in, flags, opts, longflags, longopts);
    
      opts = opts + flags;
 
      for(int i=0;i < opts.length(); i++) {
        option = opts.substring(i,i+1);
        if (c.hasOption(option)) {
          System.out.print("... found option "+option+" with value "+c.getOption(option)+"\n");
        } else {
          System.out.print("... didn't find option "+option+"\n");
        }
      }
  
      for(int i=0; i <(longopts.length+longflags.length); i++) {
        if (i < longopts.length) {
          option = longopts[i];
        } else {
          option = longflags[i - longopts.length];
        }
        if (c.hasOption(option)) {
          System.out.print("... found option "+option+" with value "+c.getOption(option)+"\n");
        } else {
          System.out.print("... didn't find option "+option+"\n");
        }
      }
 
           
      System.out.print("... Extra Parameters: ");
      for(int i=0;i < c.params.length; i++) {
        System.out.print(c.params[i]+" ");
      }
      System.out.print("\n");
       
    } catch (InvalidCommandlineArgument e) {
      System.out.println("Caught an Invalid command line argument:\n" + e.toString()+"\n");
    }   
    System.out.println("Done."); 
  }
}

